#include "model.hpp"
#include <iostream>
#include <assimp/postprocess.h>
#include <glm/gtc/type_ptr.hpp>

Model::Mesh::Mesh(const aiMesh *aiMesh, GLuint progId) {

	std::vector<glm::vec3> vertices, normals;
	std::vector<glm::vec2> texCoords;

	for (int i = 0; i < aiMesh->mNumVertices; i++) {
		if (aiMesh->HasPositions()) {
			aiVector3D v = aiMesh->mVertices[i];
			vertices.push_back(glm::vec3(v.x, v.y, v.z));
		}
		if (aiMesh->HasNormals()) {
			const aiVector3D v = aiMesh->mNormals[i];
			normals.push_back(glm::vec3(v.x, v.y, v.z));
		} else {
			normals.push_back(glm::vec3(0));
		}
		// check for texture coord set 0
		if (aiMesh->HasTextureCoords(0)) {
			const aiVector3D v = aiMesh->mTextureCoords[0][i];
			texCoords.push_back(glm::vec2(v.x, v.y));
		}
	}

	std::vector<GLuint> indices;

	for (int i = 0; i < aiMesh->mNumFaces; i++) {
		const aiFace &face = aiMesh->mFaces[i];
		if(face.mNumIndices == 3) {
			indices.push_back(face.mIndices[0]);
			indices.push_back(face.mIndices[1]);
		   	indices.push_back(face.mIndices[2]);
		}
	}

	numIndices = indices.size();
	
	glGenVertexArrays(1, &vao);
	glBindVertexArray(vao);
	
	GLuint vbos[3];
	glGenBuffers(3, vbos);
	GLuint vertexVbo = vbos[0], normalVbo = vbos[1], indicesVbo = vbos[2];
	
	GLuint posLoc = glGetAttribLocation(progId, "vPosition");
	GLuint normalLoc = glGetAttribLocation(progId, "vNormal");
	
	glBindBuffer(GL_ARRAY_BUFFER, vertexVbo);
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec3), &vertices[0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(posLoc);
	glVertexAttribPointer(posLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	glBindBuffer(GL_ARRAY_BUFFER, normalVbo);
	glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(glm::vec3), &normals[0], GL_STATIC_DRAW);
	glEnableVertexAttribArray(normalLoc);
	glVertexAttribPointer(normalLoc, 3, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indicesVbo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), &indices[0], GL_STATIC_DRAW);
};

Model::Node::Node(const aiNode &node, GLuint progId): ai(node), progId(progId) {
	for (int i = 0; i < node.mNumMeshes; i++) {
		meshIndices.push_back(node.mMeshes[i]);
	}
	for (int i = 0; i < node.mNumChildren; i++) {
		const aiNode *child = node.mChildren[i];
		children.push_back(new Node(*child, progId));
	}
}

glm::mat4 aiMatrixToMat4(aiMatrix4x4 from) {
	glm::mat4 to;
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			to[i][j] = from[j][i];
	return to;
}

void Model::Node::draw(const std::vector<Mesh> &meshes, glm::mat4 parentTrans = glm::mat4(1)) const {
	GLuint modelLoc = glGetUniformLocation(progId, "model");
	glm::mat4 m = parentTrans * aiMatrixToMat4(ai.mTransformation) * model;
	
	for (unsigned int i: meshIndices) {
		const Mesh &mesh = meshes[i];

		glBindVertexArray(mesh.vao);
		
		glUniformMatrix4fv(modelLoc, 1, GL_FALSE, glm::value_ptr(m));

		glDrawElements(GL_TRIANGLES, mesh.numIndices, GL_UNSIGNED_INT, 0);
	}
	for (Node *child: children) child->draw(meshes, m);
}

Model::Model(const std::string &path, GLuint progId): progId(progId) {
	const aiScene *scene = importer.ReadFile(path, 
			aiProcess_Triangulate |
			aiProcess_GenNormals);
	if (!scene) {
		std::cerr << importer.GetErrorString() << std::endl;
		exit(1);
	}

	for (int i = 0; i < scene->mNumMeshes; i++) {
		const aiMesh *mesh = scene->mMeshes[i];
		meshes.push_back(Mesh(mesh, progId));
	}

	for (int i = 0; i < scene->mNumMaterials; i++) {
		const aiMaterial &material = *scene->mMaterials[i];

		for (int j = 0; j < material.GetTextureCount(aiTextureType_DIFFUSE); j++) {
			aiString path;
			material.GetTexture(aiTextureType_DIFFUSE, j, &path);
		}
		std::cout << material.GetTextureCount(aiTextureType_DIFFUSE) << std::endl;
		std::cout << path.C_Str() << std::endl;
	}

	root = new Node(*(scene->mRootNode), progId);
}

void Model::draw() const {
	root->draw(meshes);
}

Model::Node* Model::find(const std::string &name) {
	const aiNode *node = root->ai.FindNode(aiString(name));
	Model::Node* res = root->findNode(*node);
	return res;
}

Model::Node* Model::Node::findNode(const aiNode &aiNode) {
	if (&ai == &aiNode) return this;
	for (Model::Node *child: children) {
		Model::Node *res = child->findNode(aiNode);
		if (res) return res;
	}
	return nullptr;
}
